#include "./FileDialogGui.h"

#include "../shared/unicode_conversion.h"
#include "../shared/it_enum_list.h"
#include "../shared/string_utilities.h"
#include "../se_sdk3/MpString.h"

using namespace gmpi;
using namespace gmpi_gui;
using namespace gmpi_sdk;
using namespace JmUnicodeConversions;

GMPI_REGISTER_GUI(MP_SUB_TYPE_GUI2, FileDialogGui, L"SE FileDialog" );

FileDialogGui::FileDialogGui() :
m_prev_trigger(false)
{
	// initialise pins.
	initializePin( 0, pinFileName,static_cast<MpGuiBaseMemberPtr2>(&FileDialogGui::OnFullPathChanged) );
	initializePin( 1,pinFileExtension);
	initializePin( 2,pinTrigger, static_cast<MpGuiBaseMemberPtr2>(&FileDialogGui::onSetTrigger) );
	initializePin( 3,pinSaveMode);
	initializePin( 4, pinName, static_cast<MpGuiBaseMemberPtr2>(&FileDialogGui::OnNameChanged) );
	initializePin( 5, pinPath);
}

void FileDialogGui::onSetTrigger()
{
	// trigger on mouse-up
	if( pinTrigger == false && m_prev_trigger == true ) // dialog triggered on mouse-up (else dialog grabs focus, button never resets)
	{
		std::wstring filename = pinFileName;
		std::wstring file_extension = pinFileExtension;

		IMpGraphicsHostBase* dialogHost = 0;
		getHost()->queryInterface(SE_IID_GRAPHICS_HOST_BASE, reinterpret_cast<void**>( &dialogHost ));

		if( dialogHost != 0 )
		{
			//gmpi_sdk::mp_shared_ptr<gmpi_gui::IMpFileDialog> returnFileDialog;
			int dialogMode = (int) pinSaveMode;
			dialogHost->createFileDialog(dialogMode, nativeFileDialog.GetAddressOf());

			if( ! nativeFileDialog.isNull() ) // returnFileDialog != 0 )
			{
				//gmpi_sdk::mp_shared_ptr<IMpFileDialog> dialog;
				//returnFileDialog->queryInterface(SE_IID_GRAPHICS_PLATFORM_FILE_DIALOG, dialog.asIMpUnknownPtr());
/*
				it_enum_list it2(pinFileExtension);
				if( it2.size() == 0 )
				{
					nativeFileDialog.AddExtension("*");
				}
				else
				{
					for( it2.First(); !it2.IsDone(); ++it2 )
					{
						nativeFileDialog.AddExtension( WStringToUtf8( (*it2)->text).c_str(), "");
					}
				}
*/
				nativeFileDialog.AddExtensionList(pinFileExtension);

				//nativeFileDialog.SetInitialFilename( (StripPath((std::string) pinFileName)).c_str() );
				//nativeFileDialog.setInitialDirectory( (StripFilename((std::string) pinFileName)).c_str() );
				nativeFileDialog.SetInitialFullPath(pinFileName);

				/*
				MpString filename;
				if( MP_OK == dialog->Show(filename.getUnknown()) )
				{
					pinFileName = Utf8ToWstring( filename.str() );
				}
				*/

				nativeFileDialog.ShowAsync([this](int32_t result) -> void { this->OnFileDialogComplete(result); });
			}
		}
	}

	m_prev_trigger = pinTrigger;
}

void FileDialogGui::OnFileDialogComplete(int32_t result)
{

	if(result == gmpi::MP_OK)
    {
    pinFileName = nativeFileDialog.GetSelectedFilename();
    OnFullPathChanged();
    }

	nativeFileDialog.setNull(); // release it.

}
void FileDialogGui::OnFullPathChanged()
{
    pinName=( (StripPath((std::string) pinFileName)).c_str() );
    //pinName.sendPinUpdate();
    pinPath = ( (StripFilename((std::string) pinFileName)).c_str() );
    //pinPath.sendPinUpdate();
}
void FileDialogGui::OnNameChanged()
{
    std::wstring name = pinName;
    std::wstring path = pinPath;

 if(name.length()>0)
 {
    path.insert(path.length(),L"\\");
    path.insert(path.length(),name);
    pinFileName=path.c_str();
 }
 else
 {
    path = L"";
    pinFileName=path.c_str();
 }


    //pinName.sendPinUpdate();
    //pinPath.sendPinUpdate();
}
